class Jena:
    def __init__(self, url, default_repository=None):

        if url.endswith("/"):
            url = url[:-1]

        self.url = url
        self.default_repository = default_repository

    def get_repository(self, specified):
        default = self.default_repository

        if all(repo == None for repo in [specified, default]):
            raise ValueError("No repository specified.")

        if specified != None:
            return specified
        else:
            return default

    def upload_file(self, file, repository=None):
        repository = self.get_repository(repository)
        return None



import sys
from os.path import normpath, dirname, abspath, join, isfile
import json
from urllib.parse import urlparse
import pickle
import traceback

from selenium import webdriver
from onepassword import OnePassword

from .domains import join_cookies, get_domain_driver, SeleniumWebsite

dir_script: str = abspath(join(normpath(dirname(__file__))))
file_cookies: str = join(dir_script, "cookies", "_.pkl")

with open(join(dir_script, "config.json"), "r") as f:
    config = json.load(f)

driver_types = {"firefox": webdriver.Firefox, "chrome": webdriver.Chrome}

domain_names = {
    "amazon": "www.amazon.com",
    "bestbuy": "www.bestbuy.com",
    "youtube": "www.youtube.com",
    "airbnb": "www.airbnb.com",
}


domain_instances = {}

onepw = None


def pop_silent(l: list, i: int = -1):
    """Pops item from list, returns None if list is empty or
    item doesn't exist at indice.

    Args:
        l (list): List to pop from.
        i (int, optional): Index to pop item from list. Defaults to -1.

    Returns:
        _type_: Item from list at index 'i'.
    """
    try:
        return l.pop(i)
    except IndexError:
        return None


def get_domain_instance(
    domain: str, driver, cookies: list[dict], args
) -> SeleniumWebsite:
    """Instantiates domain driver if doesn't exist, otherwise returns
    existing instance.

    Args:
        domain (str): Domain (e.g. 'www.amazon.com').
        driver (_type_): Selenium WebDriver.
        cookies (list[dict]): List of cookies.
        args (_type_): ArgParse arguments.

    Returns:
        SeleniumWebsite: Domain driver instance.
    """
    # Reference global variables 'domain_instances', 'config'.
    global domain_instances
    global config

    # Instantiate domain instance if not already defined.
    if domain not in domain_instances:
        # Get credentials from 1Password if specified in 'config.json'.
        credentials: dict = (
            None if args.skip_login else get_domain_credentials_1pw(domain)
        )

        # Instantiate the domain instance.
        domain_instances[domain] = get_domain_driver(
            domain, driver, credentials, cookies
        )

    return domain_instances[domain]


def scrub_urls(args):

    response = {"results": [], "errors": []}

    driver = driver_types[args.driver]()

    # Read cookies from file at location 'file_cookies'
    cookies: list = (
        None
        if args.skip_cookies or not isfile(file_cookies)
        else pickle.load(open(file_cookies, "rb"))
    )

    if args.manual_mode:
        input("\nEntering manual mode. Once complete, press enter...\n")

    # Iterate over list of urls specified.
    for url in args.url:
        try:
            # Get base domain URL.
            domain = urlparse(url).netloc

            domain_driver = get_domain_instance(domain, driver, cookies, args)

            # Scrape URL and append results to 'response' dictionary.
            response["results"].append(domain_driver.scrape_url(url))

        # If failed, do not stop program. Append message to 'response' dictionary.
        except ValueError as error:
            response["errors"].append(traceback.print_exc(error))
        except:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    # Iterate over list of search commands specified.
    for search in args.search:
        try:
            search_args = [s.strip() for s in search.split(";")]

            url = pop_silent(search_args, 0)
            search_string = pop_silent(search_args, 0)
            limit = pop_silent(search_args, 0) or 10

            if search_string == None:
                response["errors"].append(
                    "Expected search string after ';' "
                    "(e.g. '<domain>;<search_string>(;<limit>)'. "
                    f"Got '{search}'."
                )
                continue

            domain = domain_names.get(url.lower()) or urlparse(url).netloc

            domain_driver = get_domain_instance(domain, driver, cookies, args)

            # If 'domain_driver' does not have method 'search', append error.
            if not getattr(domain_driver, "search", False):
                response["errors"].append(
                    f"Domain '{domain}' does not have method 'search'."
                )
                continue

            response["results"].extend(domain_driver.search(search_string, limit))

        # If failed, do not stop program. Append message to 'response' dictionary.
        except ValueError as error:
            response["errors"].append(traceback.print_exc(error))
        except:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    # Save cookies to location 'file_cookies' if 'args.skip_cookies' not specified.
    if not args.skip_cookies:
        pickle.dump(
            join_cookies(driver.get_cookies(), cookies or []), open(file_cookies, "wb")
        )

    # Close web browser after parsing each url.
    driver.quit()

    # If no errors, delete property from response.
    if len(response["errors"]) == 0:
        del response["errors"]

    return response


def get_domain_credentials_1pw(domain):
    if domain not in config["credentials"] or config["credentials"][domain] == None:
        return None

    global onepw

    if onepw == None:
        onepw = OnePassword()

    return onepw.get_item(uuid="Amazon", fields=["username", "password"])



from test.urls import test_urls
from package.__main__ import run


def test_url():

    assert isinstance(test_urls, list)
    assert len(test_urls) > 0

    # Append urls to list of urls
    url_list = []
    for url_item in test_urls:
        if isinstance(url_item, dict):
            url = url_item["url"]
        elif isinstance(url_item, str):
            url = url_item
        url_list.append(url)

    res = run(["--url", *url_list])

    for result_item in res:
        print(result_item)



import re

from rdflib import Graph, Namespace
from rdflib.term import URIRef, BNode, Literal
from rdflib.namespace import SH, RDF, XSD

meta = Namespace("metadata:")


def get_one(l: list, description: str, allow_empty: bool = False):
    """Returns first element of list, or None if empty and 'allow_empty' is True.
    If list is not as expected, throws error with description provided.

    Args:
        l (list): List to return 1 item from.
        description (str): Description of list to be used in error message.
        allow_empty (bool, optional): Allow list to be of length 0.
            Defaults to False.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    if (allow_empty and len(l) > 1) or len(l) != 1:
        raise ValueError(f"Expected 1 {description}. Got {len(l)}.")

    if allow_empty and len(l) == 0:
        return None
    else:
        return l[0]


def json_to_graph(
    json: dict, shacl_graph: Graph, target_node_shape: URIRef, data_graph: Graph = None
) -> Graph:
    """Parse JSON into an RDF graph with respect to a sh:NodeShape definition.

    Args:
        json (dict): _description_
        shacl_graph (Graph): Graph containing sh:NodeShape definition.
        target_node_shape (URIRef): sh:NodeShape URI to use for parsing.
        data_graph (Graph, optional): Graph to parse data into. Defaults to None.

    Raises:
        ValueError: 'target_node_shape' is not defined in 'shacl_graph'.
        ValueError: Subject URI references variables that are not in JSON
        ValueError: Cannot specify both sh:datatype and sh:node on a single 
            sh:property.

    Returns:
        Graph: 'data_graph' with data from JSON added.
    """
    if data_graph == None:
        data_graph = Graph()

    # If 'target_node_shape' definition does not exist in 'shacl_graph', raise error.
    if not (len(shacl_graph.triples((target_node_shape, RDF.type, SH.NodeShape))) > 0):
        raise ValueError(
            f"Target node shape does not exist in SHACL graph: "
            + target_node_shape.n3()
        )

    # Get subject URI from 'meta:subjectURI', else use Blank Node.
    subject_uri = (
        get_one(
            shacl_graph.objects(target_node_shape, meta + "subjectURI"),
            f"meta:subjectURI for {target_node_shape.n3()}",
            allow_empty=True,
        )  # Get the object of the triple returned.
        or BNode()  # If no triples returned, use Blank Node.
    )
    # If subject URI specified on 'meta:subjectURI', see if variables are referenced.
    if type(subject_uri) == URIRef:
        uri_construct = subject_uri.n3()[1:-1]
        variables_referenced = re.findall(
            r"(?<={)[a-zA-Z][_a-zA-Z0-9]*(?=})", subject_uri
        )
        # If variables are referenced, replace with values from 'json'.
        if len(variables_referenced) > 0:
            # If 'json' does not contain any variables specified, raise error.
            if any(variable not in json for variable in variables_referenced):
                raise ValueError(
                    "Subject URI references variables that are not in JSON: '"
                    + "', '".join(filter(lambda v: v not in json, variables_referenced))
                    + "'."
                )
            # Replace variables in subject URI with values from JSON.
            for variable in variables_referenced:
                uri_construct = uri_construct.replace(f"{{{variable}}}", json[variable])
            # Replace 'subject_uri' with replaced variable string.
            subject_uri = URIRef(uri_construct)

    # Add 'rdf:type' statements to 'data_graph'.
    for type_uri in shacl_graph.objects(target_node_shape, SH.targetClass):
        data_graph.add((subject_uri, RDF.type, type_uri))

    # Add triples to 'data_graph' for each property specified.
    for property_uri in shacl_graph.objects(target_node_shape, SH.property):
        prop_predicates = {
            "path": (SH.path, True),
            "variable": (meta + "variable", True),
            # "min_count": (SH.minCount, False),
            # "max_count": (SH.maxCount, False),
            "datatype": (SH.datatype, False),
            "node": (SH.node, False),
        }
        props = {
            prop: get_one(
                shacl_graph.objects(property_uri, predicate_uri),
                description=(
                    f"predicate {predicate_uri.n3()} on sh:property "
                    f"{property_uri.n3()}, sh:NodeShape {target_node_shape.n3()}."
                ),
                allow_empty=not required,
            )
            for prop, [predicate_uri, required] in prop_predicates.items()
        }
        if props["datatype"] and props["node"]:
            raise ValueError(
                "Cannot specify both sh:datatype and sh:node on a single sh:property."
                f"Subject URI: {subject_uri.n3()}. sh:property URI: {property_uri.n3()}."
            )

        json_value_list = json[props["variable"]]
        if type(json_value_list) != list:
            json_value_list = [json_value_list]
        for value in json_value_list:
            if props["node"]:
                json_to_graph(value, shacl_graph, props["node"], data_graph)
            elif props["datatype"]:
                object_uri = (
                    URIRef(value)
                    if props["datatype"] == XSD.anyURI
                    else Literal(value, datatype=props["datatype"])
                )
                data_graph.add((subject_uri, props["path"], object_uri))
            else:
                data_graph.add((subject_uri, props["path"], Literal(value)))

    return data_graph



import sys
import json
from argparse import ArgumentParser

from .scrub import scrub_urls


def parse_args(args_list: list[str]):
    parser = ArgumentParser()

    parser.add_argument(
        "--url", "-l", nargs="+", default=[], help="URL to parse information from."
    )

    parser.add_argument(
        "--search",
        "-s",
        nargs="+",
        default=[],
        help="Search against domain (e.g. 'amazon; Hello world!').",
    )

    parser.add_argument(
        "--driver",
        "-d",
        default="firefox",
        type=str.lower,
        choices=["firefox", "chrome"],
        help="Selenium WebDriver type to use.",
    )

    parser.add_argument(
        "--skip-login",
        action="store_true",
        help="If specified, skips login on websites specified.",
    )

    parser.add_argument(
        "--skip-cookies",
        action="store_true",
        help="If specified, skips adding cookies.",
    )

    parser.add_argument(
        "--db-url",
        default="localhost:3030",
        help="URL connecting to RDF database. "
        "(Defaults to Apache Jena 'localhost:3030')",
    )

    parser.add_argument(
        "--manual-mode",
        action="store_true",
        help="If specified, puts user in an environment where they can manually "
        "enter credentials. Once finished, press enter in the terminal.",
    )

    return parser.parse_args(args_list)


def run(args_list: list[str]=None):
    if args_list is None:
        args_list = sys.argv[1:]
    
    args = parse_args(args_list)

    res = scrub_urls(args)
    
    print(json.dumps(res))
    
    return res
    
    



from setuptools import setup
from os import path, getcwd

setup(
    name="urlscrub",
    version="0.1.0",
    author="Neil Graham",
    author_email="grahamneiln@gmail.com",
    url="https://github.com/NeilGraham/urlscrub",
    license_files="LICENSE.txt",
    description="Tool for parsing URL webpage into JSON + RDF.",
    long_description=open(path.join(getcwd(), "README.md")).read()
    # Replace local links to 'docs/' to Github page 'docs/'.
    .replace(
        "](docs/",
        "](https://github.com/NeilGraham/urlscrub/blob/master/docs/",
    ),
    long_description_content_type="text/markdown",
    packages=["urlscrub"],
    package_dir={"urlscrub": "package"},
    entry_points={"console_scripts": ["urlscrub = urlscrub.__main__:run"]},
    python_requires=">=3.10",
    install_requires=[
        "pytest >= 7.1.2",
        "selenium >= 4.2.0",
        "rdfhash >= 0.3.0",
    ],
)
